import pytest
import types
import torch
import torch.optim as optim
from dingo.core.nn.nsf import (
    create_nsf_model,
    FlowWrapper,
    create_nsf_with_rb_projection_embedding_net,
)
from dingo.core.nn.enets import create_enet_with_projection_layer_and_dense_resnet
from dingo.core.utils import torchutils


@pytest.fixture()
def data_setup_nsf_large():
    d = types.SimpleNamespace()
    d.input_dim = 15
    d.context_dim = 129
    d.num_flow_steps = 30
    d.base_transform_kwargs = {
        "hidden_dim": 512,
        "num_transform_blocks": 5,
        "activation": "elu",
        "dropout_probability": 0.0,
        "batch_norm": True,
        "num_bins": 8,
        "base_transform_type": "rq-coupling",
    }
    d.embedding_net_kwargs = {
        "input_dims": (2, 3, 8033),
        # 'n_rb': 200,
        "svd": {"size": 200},
        "V_rb_list": None,
        "output_dim": 128,
        "hidden_dims": [
            1024,
            1024,
            1024,
            1024,
            1024,
            1024,
            512,
            512,
            512,
            512,
            512,
            512,
            256,
            256,
            256,
            256,
            256,
            256,
            128,
            128,
            128,
            128,
            128,
            128,
        ],
        "activation": "elu",
        "dropout": 0.0,
        "batch_norm": True,
        "added_context": True,
    }
    d.embedding_net_builder = create_enet_with_projection_layer_and_dense_resnet
    d.nde_builder = create_nsf_model
    d.nde_kwargs = {
        "input_dim": d.input_dim,
        "context_dim": d.context_dim,
        "num_flow_steps": d.num_flow_steps,
        "base_transform_kwargs": d.base_transform_kwargs,
    }
    return d


@pytest.fixture()
def data_setup_nsf_small():
    d = types.SimpleNamespace()
    d.input_dim = 4
    d.context_dim = 10
    d.num_flow_steps = 5
    d.base_transform_kwargs = {
        "hidden_dim": 64,
        "num_transform_blocks": 2,
        "activation": "elu",
        "dropout_probability": 0.0,
        "batch_norm": True,
        "num_bins": 8,
        "base_transform_type": "rq-coupling",
    }
    d.embedding_net_kwargs = {
        "input_dims": (2, 3, 20),
        # 'n_rb': 10,
        "V_rb_list": None,
        "output_dim": 8,
        "hidden_dims": [32, 16, 8],
        "activation": "elu",
        "dropout": 0.0,
        "batch_norm": True,
        "added_context": True,
        "svd": {"size": 10},
    }
    d.embedding_net_builder = create_enet_with_projection_layer_and_dense_resnet
    d.nde_builder = create_nsf_model
    d.nde_kwargs = {
        "input_dim": d.input_dim,
        "context_dim": d.context_dim,
        "num_flow_steps": d.num_flow_steps,
        "base_transform_kwargs": d.base_transform_kwargs,
    }

    d.batch_size = 20
    d.x = torch.rand((d.batch_size, *d.embedding_net_kwargs["input_dims"]))
    d.z = torch.ones(
        (d.batch_size, d.context_dim - d.embedding_net_kwargs["output_dim"])
    )
    d.y = torch.ones((d.batch_size, d.input_dim))

    # build d.yy, which depends on input d.zz
    d.xx = torch.cat((d.x, d.x))
    d.yy = torch.cat((d.y, -d.y))
    d.zz = torch.cat((d.z, -d.z))

    return d


def test_nsf_number_of_parameters(data_setup_nsf_large):
    """
    Builds a neural spline flow with the hyperparameters from that used in
    https://arxiv.org/abs/2106.12594, and checks that the number of
    parameters matches the expected one.
    """

    d = data_setup_nsf_large

    embedding_net = d.embedding_net_builder(**d.embedding_net_kwargs)
    flow = d.nde_builder(**d.nde_kwargs)
    model = FlowWrapper(flow, embedding_net)

    num_params = torchutils.get_number_of_model_parameters(model)
    assert num_params == 131448775, "Unexpected number of model parameters."


def test_sample_method_of_nsf(data_setup_nsf_small):
    """
    Test the forward pass of flow model for log_prob.
    """

    d = data_setup_nsf_small

    embedding_net = d.embedding_net_builder(**d.embedding_net_kwargs)
    flow = d.nde_builder(**d.nde_kwargs)
    model = FlowWrapper(flow, embedding_net)

    samples = model.sample(d.x, d.z)
    # model.sample(num_samples=1) adds an extra dimension that needs to be squeezed.
    samples = samples.squeeze(1)

    assert samples.shape == d.y.shape, "Unexpected shape of samples."
    assert torch.all(samples > -10) and torch.all(samples < 10), (
        "Unexpected samples encountered. Network initialization or "
        "normalization seems broken."
    )

    with pytest.raises(ValueError):
        model.sample(d.z, d.x)
    with pytest.raises(ValueError):
        model.sample(d.x, d.z, d.z)
    with pytest.raises(RuntimeError):
        model.sample(d.x, d.x)


def test_forward_pass_for_log_prob_of_nsf(data_setup_nsf_small):
    """
    Test the forward pass of flow model for log_prob.
    """

    d = data_setup_nsf_small

    embedding_net = d.embedding_net_builder(**d.embedding_net_kwargs)
    flow = d.nde_builder(**d.nde_kwargs)
    model = FlowWrapper(flow, embedding_net)

    loss = -model(d.y, d.x, d.z)
    assert list(loss.shape) == [d.batch_size], "Unexpected output shape."
    assert torch.all(loss > 0) and torch.all(loss < 40), (
        "Unexpected log prob encountered. Network initialization or "
        "normalization seems broken."
    )

    with pytest.raises(ValueError):
        model(d.y, d.z, d.x)
    with pytest.raises(ValueError):
        model(d.y, d.x, d.z, d.z)
    with pytest.raises(RuntimeError):
        model(d.y, d.x, d.x)


def test_backward_pass_for_log_prob_of_nsf(data_setup_nsf_small):
    """
    Test the backward pass of flow model for log_prob. This function also
    checks that conditional sampling improves during training.
    """

    d = data_setup_nsf_small

    embedding_net = d.embedding_net_builder(**d.embedding_net_kwargs)
    flow = d.nde_builder(**d.nde_kwargs)
    model = FlowWrapper(flow, embedding_net)
    optimizer = optim.Adam(model.parameters(), lr=0.003)

    # Simple train loop. The learned parameters yy are strongly correlated
    # with the context information d.zz. We check that (i) the loss and (ii)
    # the posterior samples improve during training.
    losses = []
    for idx in range(40):
        yy = d.yy + 0.02 * torch.rand_like(d.yy)
        xx = torch.rand_like(d.xx)
        loss = -torch.mean(model(yy, xx, d.zz))
        losses.append(loss.detach().item())
        loss.backward()
        optimizer.step()
        optimizer.zero_grad()
    samples_n = torch.mean(model.sample(d.xx, d.zz, num_samples=100), axis=1)

    assert losses[-1] < losses[0], "Loss did not improve in training."
    assert (
        torch.mean((torch.abs(samples_n - d.yy) > 0.8).float()) < 0.3
    ), "Training may not have worked. Check manually that sampling improves."


def test_model_builder_for_nsf_with_rb_embedding_net(data_setup_nsf_small):
    """
    Test the builder function create_nsf_with_rb_projection_embedding_net.
    """

    d = data_setup_nsf_small

    model = create_nsf_with_rb_projection_embedding_net(
        d.nde_kwargs, d.embedding_net_kwargs
    )

    loss = -model(d.y, d.x, d.z)
    assert list(loss.shape) == [d.batch_size], "Unexpected output shape."
    assert torch.all(loss > 0) and torch.all(loss < 40), (
        "Unexpected log prob encountered. Network initialization or "
        "normalization seems broken."
    )

    with pytest.raises(ValueError):
        model(d.y, d.z, d.x)
    with pytest.raises(ValueError):
        model(d.y, d.x, d.z, d.z)
    with pytest.raises(RuntimeError):
        model(d.y, d.x, d.x)



import pytest
import numpy as np
from scipy.stats import binom

from dingo.gw.domains import UniformFrequencyDomain, MultibandedFrequencyDomain
from dingo.gw.transforms import CropMaskStrainRandom, MaskDataForFrequencyRangeUpdate


TOLERANCE = 1e-4  # probability with which we allow confidence tests to fail


@pytest.fixture
def cropping_setup():
    domain = UniformFrequencyDomain(f_min=20, f_max=1024, delta_f=1 / 4)
    example_batch = dict(
        waveform=np.random.normal(size=(1000, 2, 3, len(domain) - domain.min_idx))
    )
    return domain, example_batch


@pytest.fixture
def cropping_setup_mfd():
    domain_settings = {
        "nodes": [20.0, 26.0, 34.0, 46.0, 62.0, 78.0, 1038.0],
        "delta_f_initial": 0.0625,
        "base_domain": {
            "type": "UniformFrequencyDomain",
            "f_min": 20.0,
            "f_max": 2048.0,
            "delta_f": 0.0625,
        },
    }
    domain = MultibandedFrequencyDomain(**domain_settings)
    example_batch = dict(
        waveform=np.random.normal(size=(1000, 2, 3, len(domain) - domain.min_idx))
    )
    return domain, example_batch


def test_cropping_init_errors(cropping_setup):
    """Check that CropMaskStrainRandom raises errors for invalid initialization inputs."""
    domain, _ = cropping_setup
    with pytest.raises(ValueError):
        _ = CropMaskStrainRandom(domain, f_min_upper=domain.f_min - 2)
    with pytest.raises(ValueError):
        _ = CropMaskStrainRandom(domain, f_max_lower=domain.f_max + 2)
    with pytest.raises(ValueError):
        _ = CropMaskStrainRandom(domain, f_min_upper=32, f_max_lower=30)


def test_cropping_frequency_calibration(cropping_setup):
    """Test that the crop frequencies are calibrated w.r.t. the strain domain."""
    domain, example_batch = cropping_setup
    strain_in = example_batch["waveform"]

    # test cropping from below
    f_min = 32
    idx_f_min = np.where(domain()[domain.min_idx :] == f_min)[0][0]
    cropping_transform = CropMaskStrainRandom(
        domain, deterministic_fmin_fmax=[f_min, None]
    )
    strain_out = cropping_transform(example_batch)["waveform"]
    # all values below 32Hz should be zero, all values above should be non-zero
    assert (strain_out[..., :idx_f_min] == 0).all()
    assert (strain_out[..., :idx_f_min] != strain_in[..., :idx_f_min]).all()
    assert (strain_out[..., idx_f_min:] != 0).all()
    assert (strain_out[..., idx_f_min:] == strain_in[..., idx_f_min:]).all()

    # test cropping from above
    f_max = 64
    idx_f_max = np.where(domain()[domain.min_idx :] == f_max)[0][0]
    cropping_transform = CropMaskStrainRandom(
        domain, deterministic_fmin_fmax=[[None, f_max], [None, f_max]]
    )
    strain_out = cropping_transform(example_batch)["waveform"]
    # all values above 64Hz should be zero, all values below should be non-zero
    assert (strain_out[..., idx_f_max + 1 :] == 0).all()
    assert (strain_out[..., idx_f_max + 1 :] != strain_in[..., idx_f_max + 1 :]).all()
    assert (strain_out[..., : idx_f_max + 1] != 0).all()
    assert (strain_out[..., : idx_f_max + 1] == strain_in[..., : idx_f_max + 1]).all()


def test_cropping_frequency_calibration_mfd(cropping_setup_mfd):
    """Test that the crop frequencies are calibrated w.r.t. the strain domain."""
    domain, example_batch = cropping_setup_mfd
    strain_in = example_batch["waveform"]

    # test cropping from below
    idx_f_min = 100
    f_min = domain()[idx_f_min]
    cropping_transform = CropMaskStrainRandom(
        domain, deterministic_fmin_fmax=[f_min, None]
    )
    strain_out = cropping_transform(example_batch)["waveform"]
    # all values below f_min should be zero, all values above should be non-zero
    assert (strain_out[..., :idx_f_min] == 0).all()
    assert (strain_out[..., :idx_f_min] != strain_in[..., :idx_f_min]).all()
    assert (strain_out[..., idx_f_min:] != 0).all()
    assert (strain_out[..., idx_f_min:] == strain_in[..., idx_f_min:]).all()

    # test cropping from above
    idx_f_max = 250
    f_max = domain()[idx_f_max]
    cropping_transform = CropMaskStrainRandom(
        domain, deterministic_fmin_fmax=[[None, f_max], [None, f_max]]
    )
    strain_out = cropping_transform(example_batch)["waveform"]
    # all values above f_max should be zero, all values below should be non-zero
    assert (strain_out[..., idx_f_max + 1 :] == 0).all()
    assert (strain_out[..., idx_f_max + 1 :] != strain_in[..., idx_f_max + 1 :]).all()
    assert (strain_out[..., : idx_f_max + 1] != 0).all()
    assert (strain_out[..., : idx_f_max + 1] == strain_in[..., : idx_f_max + 1]).all()


def test_cropping_frequency_calibration_independent_detectors(cropping_setup):
    """Test that deterministic cropping works in multiple detectors."""
    domain, example_batch = cropping_setup
    strain_in = example_batch["waveform"]

    # test cropping from below
    f_min = 32
    idx_f_min = np.where(domain()[domain.min_idx :] == f_min)[0][0]
    cropping_transform = CropMaskStrainRandom(
        domain, deterministic_fmin_fmax=[[f_min, None], [None, None]]
    )
    strain_out = cropping_transform(example_batch)["waveform"]
    # all values below 32Hz should be zero for detector 0
    # all values above should be non-zero
    assert (strain_out[:, 0, :, :idx_f_min] == 0).all()
    assert (strain_out[:, 1, :, :idx_f_min] != 0).all()

    # test that broadcasting works as intended: error should be raised when
    # len(deterministic_fmin_fmax) not in [0, num_detectors]
    cropping_transform = CropMaskStrainRandom(
        domain, deterministic_fmin_fmax=[[f_min, None], [None, None], [None, None]]
    )
    with pytest.raises(ValueError):
        _ = cropping_transform(example_batch)["waveform"]


def test_cropping_bounds(cropping_setup):
    domain, example_batch = cropping_setup
    frequencies = domain()[domain.min_idx :]
    n = 1000
    f_min_upper = domain.f_min + 2
    f_max_lower = domain.f_max - 4
    cropping_transform = CropMaskStrainRandom(
        domain,
        f_min_upper=f_min_upper,
        f_max_lower=f_max_lower,
        independent_detectors=True,
    )
    x = frequencies[None, None, :].repeat(n, 0).repeat(3, 1)
    out = cropping_transform(dict(waveform=x))["waveform"][:, 0]
    # check boundaries
    upper = np.max(out, axis=1)
    out[out == 0] = np.inf
    lower = np.min(out, axis=1)
    assert lower.min() == domain.f_min
    assert upper.max() == domain.f_max
    assert lower.max() == f_min_upper
    assert upper.min() == f_max_lower


def test_cropping_probability(cropping_setup):
    """Test that cropping probability is applied correctly."""
    domain, example_batch = cropping_setup
    frequencies = domain()[domain.min_idx :]
    strain_in = example_batch["waveform"]
    f_min = 24
    f_max = 512
    cropping_kwargs = dict(
        domain=domain,
        f_min_upper=f_min,
        f_max_lower=f_max,
        independent_detectors=True,
    )

    for p in [0.0, 0.2, 0.5, 0.8, 1.0]:
        for independent_lower_upper in [False, True]:
            cropping_transform = CropMaskStrainRandom(
                cropping_probability=p,
                independent_lower_upper=independent_lower_upper,
                **cropping_kwargs,
            )
            strain_out = cropping_transform(example_batch)["waveform"]
            is_cropped_lower = ((strain_out[..., 0]).all(axis=-1) == 0).flatten()
            is_cropped_upper = ((strain_out[..., -1]).all(axis=-1) == 0).flatten()
            n = len(is_cropped_lower)

            # probability that strain_out[..., 0] is 0 is given by p * q,
            # where q takes into account that even if the cropping transform is applied,
            # there is a chance that the cropping boundary falls on the first bin
            p_lower = p * (1 - 1 / (frequencies <= f_min).sum())
            p_upper = p * (1 - 1 / (frequencies >= f_max).sum())
            # check that correct fraction is cropped
            assert is_cropped_lower.sum() >= binom(n, p_lower).ppf(TOLERANCE)
            assert is_cropped_lower.sum() <= binom(n, p_lower).ppf(1 - TOLERANCE)
            assert is_cropped_upper.sum() >= binom(n, p_upper).ppf(TOLERANCE)
            assert is_cropped_upper.sum() <= binom(n, p_upper).ppf(1 - TOLERANCE)

            # check that cropping from below and from above is correlated iff
            # independent_detectors is True
            if 0 < p < 1:
                a = is_cropped_lower
                b = is_cropped_upper
                covariance = ((a - a.mean()) * (b - b.mean())).mean()
                correlation = covariance / (a.std() * b.std())
                # print(p, a.std(), b.std(), covariance, correlation, (a == b).mean())
                if independent_lower_upper:
                    assert np.abs(correlation) < 0.1
                elif 0.1 < p < 0.9:
                    assert np.abs(correlation) > 0.1


def test_cropping_bounds_independence(cropping_setup):
    """Check that bounds are independent along batch and optionally detector axes."""
    domain, example_batch = cropping_setup
    strain_in = example_batch["waveform"]
    example_batch_single = dict(waveform=strain_in[0])

    for independent_detectors in [False, True]:
        cropping_transform = CropMaskStrainRandom(
            domain,
            f_min_upper=32,
            f_max_lower=512,
            cropping_probability=0.8,
            independent_detectors=independent_detectors,
            independent_lower_upper=False,
        )

        # check batched version
        strain_out = cropping_transform(example_batch)["waveform"]
        mask = strain_in != strain_out
        # there should be no variation along channel axis
        assert mask.std(axis=-2).max() == 0
        # there should be variation along detector axis if independent_detectors is True
        assert (mask.std(axis=-3).max() > 0) == independent_detectors
        # there should be variation along batch
        assert mask.std(axis=-4).max() > 0

        # check non-batched version
        strain_out = cropping_transform(example_batch_single)["waveform"]
        mask = strain_in[0] != strain_out
        # there should be no variation along channel axis
        assert mask.std(axis=-2).max() == 0
        # there should be variation along detector axis if independent_detectors is True
        assert (mask.std(axis=-3).max() > 0) == independent_detectors


@pytest.fixture
def cropping_with_asd_setup():
    domain = UniformFrequencyDomain(f_min=20.0, f_max=1024.0, delta_f=1 / 8.0)
    num_f = len(domain.sample_frequencies)
    batch_size = 100
    waveform = {
        d: np.ones([batch_size, num_f], dtype=np.complex64) for d in ["H1", "L1", "V1"]
    }
    asds = {d: np.ones([batch_size, num_f]) * 1e-20 for d in ["H1", "L1", "V1"]}
    sample = {"waveform": waveform, "asds": asds}

    return domain, sample, batch_size


@pytest.fixture
def cropping_with_asd_setup_no_batch():
    domain = UniformFrequencyDomain(f_min=20.0, f_max=1024.0, delta_f=1 / 8.0)
    num_f = len(domain.sample_frequencies)
    waveform = {d: np.ones([num_f], dtype=np.complex64) for d in ["H1", "L1", "V1"]}
    asds = {d: np.ones([num_f]) * 1e-20 for d in ["H1", "L1", "V1"]}
    sample = {"waveform": waveform, "asds": asds}

    return domain, sample, 0


@pytest.fixture
def cropping_with_asd_setup_single_batch():
    domain = UniformFrequencyDomain(f_min=20.0, f_max=1024.0, delta_f=1 / 8.0)
    num_f = len(domain.sample_frequencies)
    batch_size = 1
    waveform = {
        d: np.ones([batch_size, num_f], dtype=np.complex64) for d in ["H1", "L1", "V1"]
    }
    asds = {d: np.ones([batch_size, num_f]) * 1e-20 for d in ["H1", "L1", "V1"]}
    sample = {"waveform": waveform, "asds": asds}

    return domain, sample, batch_size


@pytest.fixture
def cropping_with_asd_mfd_setup():
    domain_settings = {
        "nodes": [20.0, 26.0, 34.0, 46.0, 62.0, 78.0, 1038.0],
        "delta_f_initial": 0.0625,
        "base_domain": {
            "type": "UniformFrequencyDomain",
            "f_min": 20.0,
            "f_max": 2048.0,
            "delta_f": 0.0625,
        },
    }
    domain = MultibandedFrequencyDomain(**domain_settings)
    num_f = domain.frequency_mask_length

    batch_size = 100
    waveform = {
        d: np.ones([batch_size, num_f], dtype=np.complex64) for d in ["H1", "L1", "V1"]
    }
    asds = {d: np.ones([batch_size, num_f]) * 1e-20 for d in ["H1", "L1", "V1"]}
    sample = {"waveform": waveform, "asds": asds}

    return domain, sample, batch_size


@pytest.mark.parametrize(
    "setup",
    [
        "cropping_with_asd_setup",
        "cropping_with_asd_setup_no_batch",
        "cropping_with_asd_setup_single_batch",
        "cropping_with_asd_mfd_setup",
    ],
)
def test_MaskDataForFrequencyRangeUpdate(request, setup):
    domain, sample, batch_size = request.getfixturevalue(setup)

    frequency_update = {
        "minimum_frequency": {"H1": 30.0, "L1": 20, "V1": 30.0},
        "maximum_frequency": 1000.0,
    }

    # Initialize transform
    trafo = MaskDataForFrequencyRangeUpdate(
        domain=domain,
        minimum_frequency=frequency_update["minimum_frequency"],
        maximum_frequency=frequency_update["maximum_frequency"],
    )
    # Pass data through transform
    out = trafo(sample)

    # Check minimum_frequencies
    mask = np.logical_and(
        domain.frequency_mask,
        domain.sample_frequencies < frequency_update["minimum_frequency"]["H1"],
    )
    if batch_size > 0:
        mask = np.repeat(mask[np.newaxis, :], axis=0, repeats=batch_size)
    assert np.all(out["waveform"]["H1"][mask] == 0.0)
    assert np.all(out["asds"]["H1"][mask] == 1.0)
    mask = np.logical_and(
        domain.frequency_mask,
        domain.sample_frequencies < frequency_update["minimum_frequency"]["L1"],
    )
    if batch_size > 0:
        mask = np.repeat(mask[np.newaxis, :], axis=0, repeats=batch_size)
    assert np.all(out["waveform"]["L1"][mask] == 0.0)
    assert np.all(out["asds"]["L1"][mask] == 1.0)
    mask = np.logical_and(
        domain.frequency_mask,
        domain.sample_frequencies < frequency_update["minimum_frequency"]["V1"],
    )
    if batch_size > 0:
        mask = np.repeat(mask[np.newaxis, :], axis=0, repeats=batch_size)
    assert np.all(out["waveform"]["V1"][mask] == 0.0)
    assert np.all(out["asds"]["V1"][mask] == 1.0)

    # Check maximum frequencies
    mask = np.logical_and(
        domain.frequency_mask,
        domain.sample_frequencies > frequency_update["maximum_frequency"],
    )
    if batch_size > 0:
        mask = np.repeat(mask[np.newaxis, :], axis=0, repeats=batch_size)
    assert np.all([np.all(v[mask] == 0.0) for v in out["waveform"].values()])
    assert np.all([np.all(v[mask] == 1.0) for v in out["asds"].values()])



import numpy as np
import torchvision
import pytest

from bilby.gw.detector import InterferometerList
from bilby.core.prior import Uniform

from dingo.gw.transforms import (
    SampleExtrinsicParameters,
    GetDetectorTimes,
    GNPECoalescenceTimes,
    ProjectOntoDetectors,
    SampleNoiseASD,
    WhitenAndScaleStrain,
    AddWhiteNoiseComplex,
    SelectStandardizeRepackageParameters,
    RepackageStrainsAndASDS,
    UnpackDict,
)
from dingo.gw.prior import default_extrinsic_dict
from dingo.gw.domains import UniformFrequencyDomain
from dingo.gw.noise.asd_dataset import ASDDataset


@pytest.fixture
def standardization_dict():
    return {
        "mean": {
            "chirp_mass": 10.0,
            "mass_ratio": 0.5,
        },
        "std": {
            "chirp_mass": 1.0,
            "mass_ratio": 0.1,
        },
    }


@pytest.fixture
def domain():
    return UniformFrequencyDomain(0, 1024, 0.125)


def input_sample_batched(batch_size, input_domain):
    rand = np.random.rand(batch_size, len(input_domain)) + 1j * np.random.rand(
        batch_size, len(input_domain)
    )
    d = {
        "parameters": {
            "chirp_mass": np.random.rand(batch_size),
            "mass_ratio": np.random.rand(batch_size),
            "luminosity_distance": np.ones(batch_size) * 100.0,
            "geocent_time": np.zeros(batch_size),
        },
        "waveform": {
            "h_plus": rand,
            "h_cross": rand,
        },
    }

    return d


@pytest.fixture
def input_sample_unbatched(domain):
    rand = np.random.rand(len(domain)) + 1j * np.random.rand(len(domain)
    )
    d = {
        "parameters": {
            "chirp_mass": 12.3,
            "mass_ratio": 0.4,
            "luminosity_distance": 100.0,
            "geocent_time": 0.0,
        },
        "waveform": {
            "h_plus": rand,
            "h_cross": rand
        },
    }

    return d


@pytest.fixture
def asd_dataset(domain):
    asd = np.random.rand(5, len(domain))
    asds = {"H1": asd, "L1": asd}
    settings = {"domain_dict": domain.domain_dict}
    dictionary = {"asds": asds, "settings": settings}

    return ASDDataset(dictionary=dictionary)


@pytest.fixture
def transform_list(standardization_dict, asd_dataset, domain):
    ifo_list = InterferometerList(["H1", "L1"])
    ref_time = 1126259462.391
    transforms = [
        SampleExtrinsicParameters(default_extrinsic_dict),
        GetDetectorTimes(ifo_list, ref_time),
        GNPECoalescenceTimes(ifo_list, Uniform(minimum=-0.001, maximum=0.001), True),
        ProjectOntoDetectors(ifo_list, domain, ref_time),
        SampleNoiseASD(asd_dataset),
        WhitenAndScaleStrain(domain.noise_std),
        AddWhiteNoiseComplex(),
        SelectStandardizeRepackageParameters(
            {"inference_parameters": ["chirp_mass", "mass_ratio"]}, standardization_dict
        ),
        RepackageStrainsAndASDS(
            [ifo.name for ifo in ifo_list], first_index=domain.min_idx
        ),
        UnpackDict(["inference_parameters", "waveform"]),
    ]

    return transforms


# test that the transforms work for batched and for unbatched data
batch_sizes = [1, 2]
@pytest.mark.parametrize("batch_size", batch_sizes)
def test_batched_training_transforms(transform_list, batch_size, domain):
    input_sample = input_sample_batched(batch_size, domain)
    transforms = torchvision.transforms.Compose(transform_list)
    output = transforms(input_sample)
    assert output[0].shape[0] == batch_size
    assert output[1].shape[0] == batch_size

def test_unbatched_training_transforms(transform_list, input_sample_unbatched):
    transforms = torchvision.transforms.Compose(transform_list)
    output = transforms(input_sample_unbatched)
    assert output[0].ndim == 1
    assert output[1].ndim == 3


import argparse

import torch

from dingo.core.utils.backward_compatibility import torch_load_with_fallback


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument("--checkpoint", type=str, required=True)
    parser.add_argument("--key", type=str, required=True, nargs="+")
    parser.add_argument("--value", type=str, required=True)
    return parser.parse_args()


def main():
    args = parse_args()

    d, _ = torch_load_with_fallback(args.checkpoint)

    # Figure out type of value
    try:
        value = int(args.value)
    except ValueError:
        try:
            value = float(args.value)
        except ValueError:
            value = args.value

    settings = d["metadata"]
    for i in args.key[:-1]:
        settings = settings[i]
    settings[args.key[-1]] = value

    torch.save(d, args.checkpoint)


if __name__ == "__main__":
    main()



import pytest
from testutils_enets import *
from dingo.core.nn.enets import LinearProjectionRB, DenseResidualNet, \
    ModuleMerger, create_enet_with_projection_layer_and_dense_resnet


@pytest.fixture()
def data_setup_rb():
    batch_size = 1000
    num_blocks, num_channels, num_bins = 2, 3, 200
    n_rb = 10
    (y1, y2), (V1, V2) = \
        generate_1d_datasets_and_reduced_basis(batch_size, num_bins)
    enet_kwargs = {
        'input_dims': (num_blocks, num_channels, num_bins),
        #'n_rb': n_rb,
        'V_rb_list': (V1, V2),
        'output_dim': 8,
        'hidden_dims': [32, 16, 16, 8],
        'activation': 'elu',
        'dropout': 0.0,
        'batch_norm': True,
        'svd': {'size': n_rb}
    }
    return {
        'batch_size': batch_size,
        'num_blocks': num_blocks,
        'num_channels': num_channels,
        'num_bins': num_bins,
        'y1': y1,
        'y2': y2,
        'V1': V1,
        'V2': V2,
        'n_rb': n_rb,
        'enet_kwargs': enet_kwargs,
    }


def test_projection_of_LinearProjectionRB(data_setup_rb):
    """
    This function tests the LinearProjectionRB embedding network for the
    dimensions and the correct application of the reduced basis projection.
    """
    d = data_setup_rb
    # define projection layer
    projection_layer = LinearProjectionRB(
        input_dims=[d['num_blocks'], d['num_channels'], d['num_bins']],
        n_rb=d['n_rb'], V_rb_list=(d['V1'], d['V2']))
    # prepare data for projection_layer
    y_batch_a = get_y_batch([d['y1'], d['y2']], num_channels=d['num_channels'])
    y_batch_b = get_y_batch([np.ones_like(d['y1']), np.zeros_like(d['y1'])],
                            num_channels=d['num_channels'])
    # project onto reduced basis with projection layer
    out_a = np.array(projection_layer(y_batch_a).detach())
    out_b = np.array(projection_layer(y_batch_b).detach())
    # compute reduced basis projections with matrix multiplications as
    # comparison
    ref_a = project_onto_reduced_basis_and_concat(
        [d['y1'], d['y2']], [d['V1'], d['V2']], d['n_rb'])
    ref_b = project_onto_reduced_basis_and_concat(
        [np.ones_like(d['y1']), np.zeros_like(d['y1'])],
        [d['V1'], d['V1']], d['n_rb'])
    # check if results agree
    thr = 2e-5
    assert np.max(np.abs(out_a - ref_a)) < thr
    assert np.max(np.abs(out_b - ref_b)) < thr
    # check that results for different inputs disagree
    assert np.max(np.abs(out_a - ref_b)) > thr
    # check that channels with index >= 2 don't affect the projection
    y_batch_a[:, :, 2:, :] += torch.rand_like(y_batch_a[:, :, 2:, :])
    assert np.all(out_a == np.array(projection_layer(y_batch_a).detach())), \
        'Channels with index >= 2 should not affect rb projection.'

    # check that channels with index >= 2 do affect the projection when layer
    # is not initialized with the reduced basis
    projection_layer = LinearProjectionRB(
        input_dims=[2, d['num_channels'], d['num_bins']],
        n_rb=d['n_rb'], V_rb_list=None)
    out_a_1 = np.array(projection_layer(y_batch_a).detach())
    y_batch_a[:, :, 2:, :] -= torch.rand_like(y_batch_a[:, :, 2:, :])
    out_a_2 = np.array(projection_layer(y_batch_a).detach())
    assert np.all(out_a_1 != out_a_2), \
        'Channels with index >= 2 should affect projection layer when not ' \
        'initialized with reduced basis.'


def test_valueerrors_of_LinearProjectionRB(data_setup_rb):
    """
    Check that Error is raised if layer is initialized with inconsistent input.
    """
    d = data_setup_rb
    with pytest.raises(ValueError):
        LinearProjectionRB(
            input_dims=[2, d['num_channels'], d['num_bins']],
            n_rb=d['n_rb'], V_rb_list=(d['V1'], np.zeros_like(d['y1'])))
    with pytest.raises(ValueError):
        LinearProjectionRB(
            input_dims=[2, d['num_channels'], d['num_bins']],
            n_rb=d['n_rb'], V_rb_list=(d['V1'], d['V2'], d['V2']))
    with pytest.raises(ValueError):
        LinearProjectionRB(
            input_dims=[2, d['num_channels'], d['num_bins'] + 1],
            n_rb=d['n_rb'], V_rb_list=(d['V1'], d['V2']))
    with pytest.raises(ValueError):
        LinearProjectionRB(
            input_dims=[2, 1, d['num_bins']],
            n_rb=d['n_rb'], V_rb_list=(d['V1'], d['V2']))
    with pytest.raises(ValueError):
        LinearProjectionRB(
            input_dims=[2, d['num_channels'], d['num_bins'], 1],
            n_rb=d['n_rb'], V_rb_list=(d['V1'], d['V2']))


def test_forward_pass_of_LinearProjectionRB(data_setup_rb):
    """
    Test forward pass of the LinearProjectionRB embedding network.
    """
    d = data_setup_rb
    enet = LinearProjectionRB(
        input_dims=[d['num_blocks'], d['num_channels'], d['num_bins']],
        n_rb=d['n_rb'], V_rb_list=(d['V1'], d['V2']))
    check_model_forward_pass(
        enet, [enet.output_dim],
        (d['num_blocks'], d['num_channels'], d['num_bins']), d['batch_size'])


def test_backward_pass_of_LinearProjectionRB(data_setup_rb):
    """
    Test backward pass of the LinearProjectionRB embedding network.
    """
    d = data_setup_rb
    enet = LinearProjectionRB(
        input_dims=[d['num_blocks'], d['num_channels'], d['num_bins']],
        n_rb=d['n_rb'], V_rb_list=(d['V1'], d['V2']))
    check_model_backward_pass(
        enet, (d['num_blocks'], d['num_channels'], d['num_bins']),
        d['batch_size'])


def test_forward_pass_of_DenseResidualNet():
    """
    Test forward pass of the DenseResidualNet embedding network.
    """
    batch_size = 100
    input_dim, output_dim, hidden_dims = 120, 8, (128, 64, 32, 64, 16, 16)
    enet = DenseResidualNet(input_dim, output_dim, hidden_dims)
    check_model_forward_pass(enet, [output_dim], [input_dim], batch_size)


def test_backward_pass_of_DenseResidualNet():
    """
    Test backward pass of the DenseResidualNet embedding network.
    """
    batch_size = 100
    input_dim, output_dim, hidden_dims = 120, 8, (128, 64, 32, 64, 16, 16)
    enet = DenseResidualNet(input_dim, output_dim, hidden_dims)
    check_model_backward_pass(enet, [input_dim], batch_size)


def test_forward_pass_of_2stage_enet(data_setup_rb):
    """
    Test forward pass of the embedding network built by
    create_enet_with_projection_layer_and_dense_resnet.
    """
    d = data_setup_rb
    enet = create_enet_with_projection_layer_and_dense_resnet(
        **d['enet_kwargs'])
    check_model_forward_pass(enet, [d['enet_kwargs']['output_dim']],
                             d['enet_kwargs']['input_dims'], d['batch_size'])


def test_backward_pass_of_2stage_enet(data_setup_rb):
    """
    Test backward pass of the embedding network built by
    create_enet_with_projection_layer_and_dense_resnet.
    """
    d = data_setup_rb
    enet = create_enet_with_projection_layer_and_dense_resnet(
        **d['enet_kwargs'])
    check_model_backward_pass(enet, d['enet_kwargs']['input_dims'],
                              d['batch_size'])


def test_ModuleMerger(data_setup_rb):
    """
    Test the ModuleMerger class for correct outputs and backward passes.
    """
    d = data_setup_rb
    enet_kwargs = d['enet_kwargs']
    enet = ModuleMerger(
        (nn.Identity(),
         create_enet_with_projection_layer_and_dense_resnet(**enet_kwargs)))
    x = (torch.ones(d['batch_size'], 3),
         torch.rand(d['batch_size'], *enet_kwargs['input_dims']))

    # check backward pass and optimizer step for the model
    out_0, out_1 = check_model_backward_pass(enet, x=x)

    # check that additional context is left unchanged
    assert torch.all(out_0[:, :3] == 1), \
        'Individual embedding nets not applied correctly.'
    assert torch.all(out_1[:, :3] == 1), \
        'Individual embedding nets not applied correctly.'


def test_forward_pass_of_2stage_enet_with_context(data_setup_rb):
    """
    Test forward pass of the embedding network built by
    create_enet_with_projection_layer_and_dense_resnet with additional context.
    Check that ValueError is raised when enet is provided wrong input.
    """
    d = data_setup_rb
    enet_kwargs = d['enet_kwargs']
    enet = create_enet_with_projection_layer_and_dense_resnet(
        **enet_kwargs, added_context=True)

    # define primary and additional context
    x = torch.rand((d['batch_size'], *enet_kwargs['input_dims']))
    z = torch.ones((d['batch_size'], 2))

    check_model_forward_pass(enet, [enet_kwargs['output_dim'] + 2], x=(x, z))

    _ = enet(x, z)
    with pytest.raises(ValueError):
        enet(x)
    with pytest.raises(ValueError):
        enet((x,))


def test_backward_pass_of_2stage_enet_with_context(data_setup_rb):
    """
    Test backward pass of the embedding network built by
    create_enet_with_projection_layer_and_dense_resnet with additional context.
    """
    d = data_setup_rb
    enet_kwargs = d['enet_kwargs']
    enet = create_enet_with_projection_layer_and_dense_resnet(
        **enet_kwargs, added_context=True)

    # define primary and additional context
    x = torch.rand((d['batch_size'], *enet_kwargs['input_dims']))
    z = torch.ones((d['batch_size'], 2))

    y1, y2 = check_model_backward_pass(enet, x=(x, z))
    assert torch.all(y1[:, -2:] == 1) and torch.all(y2[:, -2:] == 1), \
        'Indentity mapping for additional context is broken.'


if __name__ == '__main__':
    pass



import pytest
import numpy as np
import torch
from bilby.core.prior import PriorDict
from bilby.gw.detector import InterferometerList

from dingo.gw.transforms import GNPECoalescenceTimes


@pytest.fixture
def gnpe_time_setup():
    kernel = "bilby.core.prior.Uniform(minimum=-0.001, maximum=0.001)"
    ifo_list = InterferometerList(["H1", "L1", "V1"])
    time_prior = PriorDict(
        {
            "geocent_time": "bilby.core.prior.Uniform(minimum=-0.1, maximum=0.1)",
            "H1_time": "bilby.core.prior.Uniform(minimum=-0.1, maximum=0.1)",
            "L1_time": "bilby.core.prior.Uniform(minimum=-0.1, maximum=0.1)",
            "V1_time": "bilby.core.prior.Uniform(minimum=-0.1, maximum=0.1)",
        }
    )
    return kernel, ifo_list, time_prior


def test_gnpe_time_training(gnpe_time_setup):
    kernel, ifo_list, time_prior = gnpe_time_setup
    ifos = [ifo.name for ifo in ifo_list]
    transform = GNPECoalescenceTimes(ifo_list, kernel, exact_global_equivariance=True)

    assert transform.context_parameters == [
        f"{ifo}_time_proxy_relative" for ifo in ifos[1:]
    ]

    # During training, the sample is assumed to be *not* batched, just consisting of an
    # array of floats.
    extrinsic_parameters = time_prior.sample()
    sample = {"extrinsic_parameters": extrinsic_parameters}
    sample_new = transform(sample)
    extrinsic_parameters_new = sample_new["extrinsic_parameters"]

    # All proxies should be included in new extrinsic parameters.
    assert {f"{ifo}_time_proxy" for ifo in ifos} <= extrinsic_parameters_new.keys()
    # Context parameters need to be in new extrinsic parameters.
    assert set(transform.context_parameters) <= extrinsic_parameters_new.keys()

    for v in extrinsic_parameters_new.values():
        assert type(v) == float  # Correct type, not an array
    for ifo in ifos:
        # Detector coalescence times should be close to 0 for data simplification.
        assert np.abs(extrinsic_parameters_new[f"{ifo}_time"]) <= 0.001

    # Check particular sum rules.
    assert np.isclose(
        extrinsic_parameters_new["geocent_time"],
        extrinsic_parameters["geocent_time"]
        - extrinsic_parameters[f"{ifos[0]}_time"]
        + extrinsic_parameters_new[f"{ifos[0]}_time"],
    )
    for ifo in ifos[1:]:
        assert np.isclose(
            extrinsic_parameters_new[f"{ifo}_time_proxy_relative"],
            extrinsic_parameters[f"{ifo}_time"]
            - extrinsic_parameters_new[f"{ifo}_time"]
            - extrinsic_parameters[f"{ifos[0]}_time"]
            + extrinsic_parameters_new[f"{ifos[0]}_time"],
        )


def test_gnpe_time_inference(gnpe_time_setup):
    kernel, ifo_list, time_prior = gnpe_time_setup
    ifos = [ifo.name for ifo in ifo_list]
    transform = GNPECoalescenceTimes(
        ifo_list, kernel, exact_global_equivariance=True, inference=True
    )

    # For inference, the sample is assumed to be a torch tensor and batched.
    batch_size = 10
    extrinsic_parameters = time_prior.sample(batch_size)
    extrinsic_parameters.pop("geocent_time")  # Should not be included at this point.
    extrinsic_parameters = {k: torch.tensor(v) for k, v in extrinsic_parameters.items()}
    sample = {"extrinsic_parameters": extrinsic_parameters}
    sample_new = transform(sample)
    extrinsic_parameters_new = sample_new["extrinsic_parameters"]

    # All proxies should be included in new extrinsic parameters.
    assert {f"{ifo}_time_proxy" for ifo in ifos} <= extrinsic_parameters_new.keys()
    # Context parameters need to be in new extrinsic parameters.
    assert set(transform.context_parameters) <= extrinsic_parameters_new.keys()

    for v in extrinsic_parameters_new.values():
        assert type(v) == torch.Tensor and len(v) == batch_size

    # Check particular quantities differ by <= epsilon.
    for ifo in ifos:
        assert torch.allclose(
            extrinsic_parameters_new[f"{ifo}_time"],
            -extrinsic_parameters[f"{ifo}_time"],
            atol=0.001,
        )

    # Check particular equalities.
    for ifo in ifos[1:]:
        assert torch.equal(
            extrinsic_parameters_new[f"{ifo}_time_proxy_relative"],
            -extrinsic_parameters_new[f"{ifo}_time"]
            + extrinsic_parameters_new[f"{ifos[0]}_time"],
        )
    assert torch.equal(
        extrinsic_parameters_new["geocent_time"],
        extrinsic_parameters_new[f"{ifos[0]}_time"],
    )



import os

from bilby_pipe.job_creation.nodes import PlotNode as BilbyPlotNode

from dingo.pipe.utils import _strip_unwanted_submission_keys


class PlotNode(BilbyPlotNode):

    run_node_on_osg = False
    
    def __init__(self, inputs, merged_node, dag):
        # Call Node.__init__ directly, skipping BilbyPlotNode.__init__ to customize
        super(BilbyPlotNode, self).__init__(inputs)
        self.dag = dag
        self.job_name = merged_node.job_name.replace("_merge", "") + "_plot"
        self.label = merged_node.job_name.replace("_merge", "") + "_plot"
        self.request_cpus = 1

        # Add file transfer logic (copied from BilbyPlotNode)
        if self.inputs.transfer_files or self.inputs.osg:
            input_files_to_transfer = [
                self._relative_topdir(merged_node.result_file, self.inputs.initialdir),
            ] + inputs.additional_transfer_paths
            if self.transfer_container:
                input_files_to_transfer.append(self.inputs.container)

            input_files_to_transfer, need_scitokens = self.job_needs_authentication(
                input_files_to_transfer
            )
            if need_scitokens:
                self.extra_lines.extend(self.scitoken_lines)

            self.extra_lines.extend(
                self._condor_file_transfer_lines(
                    input_files_to_transfer,
                    [self._relative_topdir(self.inputs.outdir, self.inputs.initialdir)],
                )
            )

        self.setup_arguments(
            add_ini=False, add_unknown_args=False, add_command_line_args=False
        )
        self.arguments.add("label", self.label)
        if self.inputs.transfer_files or self.inputs.osg:
            self.arguments.add("result", os.path.relpath(merged_node.result_file))
            self.arguments.add("outdir", os.path.relpath(self.inputs.result_directory))
        else:
            self.arguments.add("result", merged_node.result_file)
            self.arguments.add("outdir", self.inputs.result_directory)
        for plot_type in ["corner", "weights", "log_probs"]:
            if getattr(inputs, f"plot_{plot_type}", False):
                self.arguments.add_flag(plot_type)
        # self.arguments.add("format", inputs.plot_format)

        if getattr(self, "disable_hdf5_locking", None):
            self.extra_lines.append('environment = "HDF5_USE_FILE_LOCKING=FALSE"')

        self.process_node()
        self.job.add_parent(merged_node.job)

        if self.inputs.simple_submission:
            _strip_unwanted_submission_keys(self.job)

    @property
    def executable(self):
        return self._get_executable_path("dingo_pipe_plot")


class PlotPPNode(BilbyPlotNode):
    def __init__(self, inputs, merged_node_list, dag):
        super(BilbyPlotNode, self).__init__(inputs)
        self.dag = dag
        self.request_cpus = 1
        self.job_name = f"{self.inputs.label}_plot_pp"
        self.setup_arguments(
            add_ini=False, add_unknown_args=False, add_command_line_args=False
        )
        self.arguments.add_positional_argument(self.inputs.result_directory)

        if getattr(self, "disable_hdf5_locking", None):
            self.extra_lines.append('environment = "HDF5_USE_FILE_LOCKING=FALSE"')

        self.process_node()
        for node in merged_node_list:
            self.job.add_parent(node.job)

        if self.inputs.simple_submission:
            _strip_unwanted_submission_keys(self.job)

    @property
    def executable(self):
        return self._get_executable_path("dingo_pipe_pp_test")

    @property
    def request_memory(self):
        return "16 GB"


import numpy as np
import torch
import pandas as pd
from dingo.gw.prior import BBHExtrinsicPriorDict
from .utils import get_batch_size_of_input_sample


class SampleExtrinsicParameters(object):
    """
    Sample extrinsic parameters and add them to sample in a separate dictionary.
    """

    def __init__(self, extrinsic_prior_dict):
        self.extrinsic_prior_dict = extrinsic_prior_dict
        self.prior = BBHExtrinsicPriorDict(extrinsic_prior_dict)

    def __call__(self, input_sample):
        sample = input_sample.copy()
        batched, batch_size = get_batch_size_of_input_sample(input_sample)
        extrinsic_parameters = self.prior.sample(batch_size if batched else None)
        extrinsic_parameters = {
            k: v.astype(np.float32) if batched else float(v)
            for k, v in extrinsic_parameters.items()
        }
        sample["extrinsic_parameters"] = extrinsic_parameters
        return sample

    @property
    def reproduction_dict(self):
        return {"extrinsic_prior_dict": self.extrinsic_prior_dict}


class SelectStandardizeRepackageParameters(object):
    """
    This transformation selects the parameters in standardization_dict,
    normalizes them by setting p = (p - mean) / std, and repackages the
    selected parameters to a numpy array.

    as_type: str = None
        only applies, if self.inverse == True
        * if None, data type is kept
        * if 'dict', dict with
        * if 'pandas', use pandas.DataFrame
    """

    def __init__(
        self,
        parameters_dict,
        standardization_dict,
        inverse=False,
        as_type=None,
        device="cpu",
    ):
        self.parameters_dict = parameters_dict
        self.mean = standardization_dict["mean"]
        self.std = standardization_dict["std"]
        self.N = len(self.mean.keys())
        if self.mean.keys() != self.std.keys():
            raise ValueError("Keys of means and stds do not match.")
        self.inverse = inverse
        self.as_type = as_type
        self.device = device

    def __call__(self, input_sample, as_type=None):
        """
        * if self.inverse == False:
            Normalize parameters (specified in self.parameters_dict),
            repackage to numpy array.
        * if self.inverse == True:
            Applies only to sample['inference_parameters'].
            Undo normalization and return as type self.as_type.
            Also transform input_sample['log_prob'], if present, according to the
            change-of-variables rule.

        Parameters
        ----------
        input_sample: dict
            input sample

        Returns
        -------
        sample: dict
            transformed sample
        """

        if not self.inverse:
            # Look for parameters in either the parameters dict, or the
            # extrinsic_parameters dict. extrinsic_parameters supersedes.
            if "extrinsic_parameters" in input_sample:
                full_parameters = {
                    **input_sample["parameters"],
                    **input_sample["extrinsic_parameters"],
                }
            else:
                full_parameters = input_sample["parameters"]

            sample = input_sample.copy()
            for k, v in self.parameters_dict.items():
                if len(v) > 0:
                    if isinstance(full_parameters[v[0]], torch.Tensor):
                        standardized = torch.empty(
                            (*full_parameters[v[0]].shape, len(v)),
                            dtype=torch.float32,
                            device=self.device,
                        )
                    elif isinstance(full_parameters[v[0]], np.ndarray):
                        standardized = np.empty(
                            (*full_parameters[v[0]].shape, len(v)), dtype=np.float32
                        )
                    else:
                        standardized = np.empty(len(v), dtype=np.float32)
                    for idx, par in enumerate(v):
                        if self.std[par] == 0:
                            raise ValueError(
                                f"Parameter {par} with standard deviation zero is included in inference parameters. "
                                f"This is not allowed. Please remove it from inference_parameters or create a new "
                                f"dataset where std({par}) is not zero."
                            )
                        standardized[..., idx] = (
                            full_parameters[par] - self.mean[par]
                        ) / self.std[par]
                    sample[k] = standardized

        else:
            sample = input_sample.copy()
            inference_parameters = self.parameters_dict["inference_parameters"]

            parameters = input_sample["parameters"][:]
            assert parameters.shape[-1] == len(inference_parameters), (
                f"Expected {len(inference_parameters)} parameters "
                f"({inference_parameters}), but got {parameters.shape[-1]}."
            )

            # de-normalize parameters
            for idx, par in enumerate(inference_parameters):
                parameters[..., idx] = (
                    parameters[..., idx] * self.std[par] + self.mean[par]
                )

            # TODO: Can we remove the as_type option? Do we ever want anything other
            #  than a dict?
            # return normalized parameters as desired type
            if self.as_type is None:
                sample["parameters"] = parameters

            elif self.as_type == "dict":
                sample["parameters"] = {}
                for idx, par in enumerate(inference_parameters):
                    sample["parameters"][par] = parameters[..., idx]

            elif self.as_type == "pandas":
                sample["parameters"] = pd.DataFrame(
                    np.array(parameters), columns=inference_parameters
                )

            else:
                raise NotImplementedError(
                    f"Unexpected type {self.as_type}, "
                    f"expected one of [None, pandas, dict]."
                )

            # TODO: Implement this for the forward map, if needed.
            if "log_prob" in sample:
                log_std = np.sum(np.log([self.std[p] for p in inference_parameters]))
                sample["log_prob"] -= log_std

        return sample


class StandardizeParameters:
    """
    Standardize parameters according to the transform (x - mu) / std.
    """

    def __init__(self, mu, std):
        """
        Initialize the standardization transform with means
        and standard deviations for each parameter

        Parameters
        ----------
        mu : Dict[str, float]
            The (estimated) means
        std : Dict[str, float]
            The (estimated) standard deviations
        """
        self.mu = mu
        self.std = std
        if not set(mu.keys()) == set(std.keys()):
            raise ValueError(
                "The keys in mu and std disagree:" f"mu: {mu.keys()}, std: {std.keys()}"
            )

    def __call__(self, samples):
        """Standardize the parameter array according to the
        specified means and standard deviations.

        Parameters
        ----------
        samples: Dict[Dict, Dict]
            A nested dictionary with keys 'parameters', 'waveform',
            'noise_summary'.

        Only parameters included in mu, std get transformed.
        """
        x = samples["parameters"]
        y = {k: (x[k] - self.mu[k]) / self.std[k] for k in self.mu.keys()}
        samples_out = samples.copy()
        samples_out["parameters"] = y
        return samples_out

    def inverse(self, samples):
        """De-standardize the parameter array according to the
        specified means and standard deviations.

        Parameters
        ----------
        samples: Dict[Dict, Dict]
            A nested dictionary with keys 'parameters', 'waveform',
            'noise_summary'.

        Only parameters included in mu, std get transformed.
        """
        y = samples["parameters"]
        x = {k: self.mu[k] + y[k] * self.std[k] for k in self.mu.keys()}
        samples_out = samples.copy()
        samples_out["parameters"] = x
        return samples_out



import argparse
from pathlib import Path

from bilby_pipe.utils import logger

from dingo.core.result import make_pp_plot
from dingo.gw.result import Result

logger.name = "dingo_pipe"


def create_parser():
    parser = argparse.ArgumentParser(
        prog="dingo_pipe PP test",
        usage="Generates a pp plot from a directory containing a set of results",
    )
    parser.add_argument("directory", help="Path to the result files")
    parser.add_argument(
        "--outdir", help="Path to output directory, defaults to input directory "
    )
    parser.add_argument("--label", help="Additional label to use for output")
    parser.add_argument(
        "--print", action="store_true", help="Print the list of filenames used"
    )
    parser.add_argument(
        "-n", type=int, help="Number of samples to truncate to", default=None
    )
    parser.add_argument(
        "--filter",
        type=str,
        help="A string to match and filtering results. If not specified, will look "
        "first for '*importance_sampling.hdf5', then '*sampling.hdf5'.",
        default=None,
    )
    return parser


def get_results_filenames(args):
    p = Path(args.directory)
    if args.filter:
        results_files = list(p.glob(f"*{args.filter}*.hdf5"))
    else:
        results_files = list(p.glob("*importance_sampling.hdf5"))
        if len(results_files) == 0:
            results_files = list(p.glob("*sampling.hdf5"))
    if len(results_files) == 0:
        raise FileNotFoundError(f"No results found in path {args.directory}")

    if args.n is not None:
        logger.info(f"Truncating to first {args.n} results")
        results_files = results_files[: args.n]
    return results_files


def get_basename(args):
    if args.outdir is None:
        args.outdir = args.directory
    basename = f"{args.outdir}/"
    if args.label is not None:
        basename += f"{args.label}_"
    return basename


def main(args=None):
    if args is None:
        args = create_parser().parse_args()
    results_filenames = get_results_filenames(args)
    results = [Result(file_name=f) for f in results_filenames]
    basename = get_basename(args)

    logger.info("Generating PP plot.")
    # keys = [
    #     name
    #     for name, p in results[0].prior.items()
    #     if isinstance(p, str) or p.is_fixed is False
    # ]
    logger.info(f"Parameters = {results[0].search_parameter_keys}")
    make_pp_plot(results, filename=f"{basename}pp.pdf")
    if "weights" in results[0].samples:
        logger.info("Generating IS PP plot.")
        make_pp_plot(results, filename=f"{basename}pp_IS.pdf", weighted=True)



import copy

import torch

from dingo.core.utils import build_train_and_test_loaders
from dingo.core.utils.trainutils import RuntimeLimits
import numpy as np
import argparse

from dingo.core.posterior_models import NormalizingFlowPosteriorModel


class SampleDataset(torch.utils.data.Dataset):
    """
    Dataset class for unconditional density estimation.
    This is required, since the training method of dingo.core.posterior_models.Base
    expects a tuple of (theta, *context) as output of the DataLoader, but here we have
    no context, so len(context) = 0. This SampleDataset therefore returns a tuple
    (theta, ) instead of just theta.
    """

    def __init__(self, data):
        self.data = data
        self.length = len(self.data)

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        """Return the data and labels at the given index as a tuple of length 1."""
        return (self.data[index],)


def train_unconditional_density_estimator(
    result,
    settings: dict,
    train_dir: str,
):
    """
    Train unconditional density estimator for a given set of samples.

    Parameters
    ----------
    result: Result
        Contains the samples on which to train the density estimator.
    settings: dict
        Dictionary containing the settings for the density estimator.
    train_dir: str
        Path to the directory where the trained model should be saved.

    Returns
    -------
    model: NormalizingFlowPosteriorModel
        trained density estimator
    """
    samples = result.samples
    # Process samples: select parameters, normalize, and convert to torch tensor
    if "parameters" in settings["data"] and settings["data"]["parameters"]:
        parameters = settings["data"]["parameters"]
    else:
        parameters = list(samples.keys())
    samples = np.array(samples[parameters])
    num_samples, num_params = samples.shape
    mean, std = np.mean(samples, axis=0), np.std(samples, axis=0)
    settings["data"]["standardization"] = {
        "mean": {param: mean[i].item() for i, param in enumerate(parameters)},
        "std": {param: std[i].item() for i, param in enumerate(parameters)},
    }
    settings["data"]["unconditional"] = True
    # normalized torch samples
    samples_torch = torch.from_numpy((samples - mean) / std).float()

    # set up density estimation network
    settings["model"]["posterior_kwargs"]["input_dim"] = num_params
    settings["model"]["posterior_kwargs"]["context_dim"] = None
    # TODO: Allow for other types of density estimators (e.g., flow matching).
    model = NormalizingFlowPosteriorModel(
        metadata={"train_settings": settings, "base": copy.deepcopy(result.metadata)},
        device=settings["training"]["device"],
    )
    model.optimizer_kwargs = settings["training"]["optimizer"]
    model.scheduler_kwargs = settings["training"]["scheduler"]
    model.initialize_optimizer_and_scheduler()

    # Store context and event metadata to keep a record, even though it will not be used
    # in training.
    model.context = result.context
    model.event_metadata = result.event_metadata

    # set up dataloaders
    train_loader, test_loader = build_train_and_test_loaders(
        SampleDataset(samples_torch),
        settings["training"]["train_fraction"],
        settings["training"]["batch_size"],
        settings["training"]["num_workers"],
    )

    # train model
    runtime_limits = RuntimeLimits(
        epoch_start=0,
        max_epochs_total=settings["training"]["epochs"],
    )
    model.train(
        train_loader,
        test_loader,
        train_dir=train_dir,
        runtime_limits=runtime_limits,
    )

    return model


def parse_args():
    parser = argparse.ArgumentParser(
        description="Unconditional density estimation for dingo samples."
    )
    parser.add_argument(
        "--settings",
        type=str,
        required=True,
        help="Path to settings file.",
    )
    return parser.parse_args()



import numpy as np


def save_training_injection(outname, pm, data, idx=0):
    """
    For debugging: extract a training injection. To be used inside train or test loop.
    """
    param_names = pm.metadata["train_settings"]["data"]["inference_parameters"]
    mean = pm.metadata["train_settings"]["data"]["standardization"]["mean"]
    std = pm.metadata["train_settings"]["data"]["standardization"]["std"]
    params = {p: data[0][idx, idx_p] for idx_p, p in enumerate(param_names)}
    params = {p: float(v * std[p] + mean[p]) for p, v in params.items()}

    from dingo.gw.domains import build_domain_from_model_metadata

    domain = build_domain_from_model_metadata(pm.metadata)
    detectors = pm.metadata["train_settings"]["data"]["detectors"]
    d = np.array(data[1])
    asds = {
        ifo: 1 / d[idx, idx_ifo, 2] * 1e-23 for idx_ifo, ifo in enumerate(detectors)
    }
    strains = {
        ifo: (d[idx, idx_ifo, 0] + 1j * d[idx, idx_ifo, 1])
        * (asds[ifo] * domain.noise_std)
        for idx_ifo, ifo in enumerate(detectors)
    }

    out_data = {"parameters": params, "asds": asds, "strains": strains}
    np.save(outname, out_data)

    from dingo.gw.injection import GWSignal

    signal = GWSignal(
        pm.metadata["dataset_settings"]["waveform_generator"],
        domain,
        domain,
        pm.metadata["train_settings"]["data"]["detectors"],
        pm.metadata["train_settings"]["data"]["ref_time"],
    )
    params_2 = params.copy()
    params_2["phase"] = (params_2["phase"] + np.pi / 2.0) % (2 * np.pi)
    params_3 = {p: v * 0.99 for p, v in params.items()}
    sample = signal.signal(params)
    sample_2 = signal.signal(params_2)
    sample_3 = signal.signal(params_3)

    import matplotlib.pyplot as plt

    plt.plot(np.abs(sample["waveform"]["H1"])[domain.min_idx :])
    plt.plot(np.abs(strains["H1"]), lw=0.8)
    plt.show()

    plt.plot(sample["waveform"]["H1"][domain.min_idx :])
    plt.plot(strains["H1"], lw=0.8)
    # plt.plot(sample_2["waveform"]["H1"][domain.min_idx:])
    plt.show()



import argparse

import torch
import yaml

from dingo.core.utils.backward_compatibility import torch_load_with_fallback


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument("--checkpoint", type=str, required=True)
    parser.add_argument("--out_file", type=str, required=True)
    return parser.parse_args()


def main():
    args = parse_args()

    d, _ = torch_load_with_fallback(args.checkpoint)

    data_settings = d["metadata"]["train_settings"]["data"]

    if "gnpe_time_shifts" in data_settings:

        data_settings["gnpe_time_shifts"][
            "kernel"
        ] = "bilby.core.prior.Uniform(minimum=-0.001, maximum=0.001)"
        del data_settings["gnpe_time_shifts"]["kernel_kwargs"]

        data_settings["context_parameters"] = [
            f"{ifo}_time_proxy" for ifo in data_settings["detectors"][1:]
        ]

        for p in data_settings["context_parameters"]:
            data_settings["standardization"]["mean"][p] = 0.0
            data_settings["standardization"]["std"][p] = data_settings[
                "standardization"
            ]["std"]["geocent_time"]

    data_settings["inference_parameters"] = data_settings["selected_parameters"]
    del data_settings["selected_parameters"]

    print("New data_settings:")
    print(yaml.dump(data_settings, default_flow_style=False, sort_keys=False))

    torch.save(d, args.out_file)


if __name__ == "__main__":
    main()



from dingo.core.dataset import DingoDataset


class EventDataset(DingoDataset):
    """Dataset class for storing single event."""

    dataset_type = "event_dataset"

    def __init__(self, file_name=None, dictionary=None):
        self.data = None
        self.settings = None
        super().__init__(
            file_name=file_name,
            dictionary=dictionary,
            data_keys=["data"],
        )



"""
This module contains the abstract base class for representing posterior models,
as well as functions for training and testing across an epoch.
"""

from abc import abstractmethod, ABC
import os
from os.path import join
import h5py

import torch
import dingo.core.utils as utils
from torch.utils.data import Dataset
import time
import numpy as np
from threadpoolctl import threadpool_limits
import dingo.core.utils.trainutils
import json
from collections import OrderedDict
from typing import Optional
from dingo.core.utils.backward_compatibility import update_model_config
from dingo.core.utils.misc import get_version

from dingo.core.utils.trainutils import EarlyStopping


class BasePosteriorModel(ABC):
    """
    Abstract base class for PosteriorModels. This is intended to construct and hold a
    neural network for estimating the posterior density, as well as saving / loading,
    and training.

    Subclasses must implement methods for constructing the specific network, sampling,
    density evaluation, and computing the loss during training.
    """

    def __init__(
        self,
        model_filename: str = None,
        metadata: dict = None,
        initial_weights: dict = None,
        device: str = "cuda",
        load_training_info: bool = True,
    ):
        """
        Initialize a model for the posterior distribution.

        Parameters
        ----------
        model_filename: str
            If given, loads data from the given file.
        metadata: dict
            If given, initializes the model from these settings
        initial_weights: dict
            Initial weights for the model
        device: str
        load_training_info: bool
        """

        self.version = f"dingo={get_version()}"  # dingo version

        self.device = None
        self.optimizer_kwargs = None
        self.network_kwargs = None
        self.scheduler_kwargs = None
        self.initial_weights = initial_weights

        self.metadata = metadata
        if self.metadata is not None:
            self.model_kwargs = self.metadata["train_settings"]["model"]
            # Expect self.optimizer_settings and self.scheduler_settings to be set
            # separately, and before calling initialize_optimizer_and_scheduler().

        self.epoch = 0
        self.network = None
        self.optimizer = None
        self.scheduler = None
        self.context = None
        self.event_metadata = None

        # build model
        if model_filename is not None:
            self.load_model(
                model_filename, load_training_info=load_training_info, device=device
            )
        else:
            self.initialize_network()
            self.network_to_device(device)

    @abstractmethod
    def initialize_network(self):
        """
        Initialize the network backbone for the posterior model.
        """
        pass

    @abstractmethod
    def sample(self, *context: torch.Tensor, num_samples: int = 1):
        """
        Sample parameters theta from the posterior model,

        theta ~ p(theta | context)

        Parameters
        ----------
        context: torch.Tensor
            Context information (typically observed data). Should have a batch
            dimension (even if size B = 1).
        num_samples: int = 1
            Number of samples to generate.

        Returns
        -------
        samples: torch.Tensor
            Shape (B, num_samples, dim(theta))
        """
        pass

    @abstractmethod
    def sample_and_log_prob(self, *context: torch.Tensor, num_samples: int = 1):
        """
        Sample parameters theta from the posterior model,

        theta ~ p(theta | context)

        and also return the log_prob. For models such as normalizing flows, it is more
        economical to calculate the log_prob at the same time as sampling, rather than
        as a separate step.

        Parameters
        ----------
        context: torch.Tensor
            Context information (typically observed data). Should have a batch
            dimension (even if size B = 1).
        num_samples: int = 1
            Number of samples to generate.

        Returns
        -------
        samples, log_prob: torch.Tensor, torch.Tensor
            Shapes (B, num_samples, dim(theta)), (B, num_samples)
        """
        pass

    @abstractmethod
    def log_prob(self, theta: torch.Tensor, *context: torch.Tensor):
        """
        Evaluate the log posterior density,

        log p(theta | context)

        Parameters
        ----------
        theta: torch.Tensor
            Parameter values at which to evaluate the density. Should have a batch
            dimension (even if size B = 1).
        context: torch.Tensor
            Context information (typically observed data). Must have context.shape[0] = B.

        Returns
        -------
        log_prob: torch.Tensor
            Shape (B,)
        """
        pass

    @abstractmethod
    def loss(self, theta: torch.Tensor, *context: torch.Tensor):
        """
        Compute the loss for a batch of data.

        Parameters
        ----------
        theta: torch.Tensor
            Parameter values at which to evaluate the density. Should have a batch
            dimension (even if size B = 1).
        context: torch.Tensor
            Context information (typically observed data). Must have the same leading
            (batch) dimension as theta.

        Returns
        -------
        loss: torch.Tensor
            Mean loss across the batch (a scalar).
        """
        pass

    def network_to_device(self, device):
        """
        Put model to device, and set self.device accordingly.
        """
        if device not in ("cpu", "cuda"):
            raise ValueError(f"Device should be either cpu or cuda, got {device}.")
        self.device = torch.device(device)
        # Commented below so that code runs on first cuda device in the case of multiple.
        # if device == 'cuda' and torch.cuda.device_count() > 1:
        #     print("Using", torch.cuda.device_count(), "GPUs.")
        #     raise NotImplementedError('This needs testing!')
        #     # dim = 0 [512, ...] -> [256, ...], [256, ...] on 2 GPUs
        #     self.network = torch.nn.DataParallel(self.network)
        print(f"Putting posterior model to device {self.device}.")
        self.network.to(self.device)

    def initialize_optimizer_and_scheduler(self):
        """
        Initializes the optimizer and scheduler with self.optimizer_kwargs
        and self.scheduler_kwargs, respectively.
        """
        if self.optimizer_kwargs is not None:
            self.optimizer = utils.get_optimizer_from_kwargs(
                self.network.parameters(), **self.optimizer_kwargs
            )
        if self.scheduler_kwargs is not None:
            self.scheduler = utils.get_scheduler_from_kwargs(
                self.optimizer, **self.scheduler_kwargs
            )

    def save_model(
        self,
        model_filename: str,
        save_training_info: bool = True,
    ):
        """
        Save the posterior model to the disk.

        Parameters
        ----------
        model_filename: str
            filename for saving the model
        save_training_info: bool
            specifies whether information required to proceed with training is
            saved, e.g. optimizer state dict

        """
        model_dict = {
            "model_kwargs": self.model_kwargs,
            "model_state_dict": self.network.state_dict(),
            "epoch": self.epoch,
            "version": self.version,
        }

        if self.metadata is not None:
            model_dict["metadata"] = self.metadata

        if self.context is not None:
            model_dict["context"] = self.context

        if self.event_metadata is not None:
            model_dict["event_metadata"] = self.event_metadata

        if save_training_info:
            model_dict["optimizer_kwargs"] = self.optimizer_kwargs
            model_dict["scheduler_kwargs"] = self.scheduler_kwargs
            if self.optimizer is not None:
                model_dict["optimizer_state_dict"] = self.optimizer.state_dict()
            if self.scheduler is not None:
                model_dict["scheduler_state_dict"] = self.scheduler.state_dict()

        torch.save(model_dict, model_filename)

    def _load_model_from_hdf5(self, model_filename: str):
        """
        Helper function to load a trained model that has been
        saved in HDF5 format using `dingo_pt_to_hdf5`.
        Parameters
        ----------
        model_filename: str
            path to saved model; must have extension '.hdf5'
        Returns
        -------
        d: dict
            A stripped down version of the dict saved by torch.save()
            Specifically, it does not include 'optimizer_state_dict'
            to save space at inference time.
        """
        d = {}
        with h5py.File(model_filename, "r") as fp:
            model_basename = os.path.basename(model_filename)
            if fp.attrs["CANONICAL_FILE_BASENAME"] != model_basename:
                raise ValueError(
                    "HDF5 attribute CANONICAL_FILE_BASENAME differs from model name",
                    model_basename,
                )

            # Load small nested dicts from json
            for k, v in fp["serialized_dicts"].items():
                d[k] = json.loads(v[()])

            # Load model weights
            model_state_dict = OrderedDict()
            for k, v in fp["model_weights"].items():
                model_state_dict[k] = torch.from_numpy(np.array(v, dtype=np.float32))
            d["model_state_dict"] = model_state_dict

        return d

    def load_model(
        self,
        model_filename: str,
        load_training_info: bool = True,
        device: str = "cuda",
    ):
        """
        Load a posterior model from the disk.

        Parameters
        ----------
        model_filename: str
            path to saved model
        load_training_info: bool #TODO: load information for training
            specifies whether information required to proceed with training is
            loaded, e.g. optimizer state dict
        device: str
        """

        # Make sure that when the model is loaded, the torch tensors are put on the
        # device indicated in the saved metadata. External routines run on a cpu
        # machine may have moved the model from 'cuda' to 'cpu'.
        ext = os.path.splitext(model_filename)[-1]
        if ext == ".pt":
            d = torch.load(model_filename, map_location=device)
        elif ext == ".hdf5":
            d = self._load_model_from_hdf5(model_filename)
        else:
            raise ValueError("Models should be ether in .pt or .hdf5 format.")

        self.version = d.get("version")

        self.model_kwargs = d["model_kwargs"]
        update_model_config(self.model_kwargs)  # For backward compatibility
        self.initialize_network()

        self.epoch = d["epoch"]

        self.metadata = d["metadata"]

        if "context" in d:
            self.context = d["context"]

        if "event_metadata" in d:
            self.event_metadata = d["event_metadata"]

        if device != "meta":
            self.network.load_state_dict(d["model_state_dict"])

            self.network_to_device(device)

            if load_training_info:
                if "optimizer_kwargs" in d:
                    self.optimizer_kwargs = d["optimizer_kwargs"]
                if "scheduler_kwargs" in d:
                    self.scheduler_kwargs = d["scheduler_kwargs"]
                # initialize optimizer and scheduler
                self.initialize_optimizer_and_scheduler()
                # load optimizer and scheduler state dict
                if "optimizer_state_dict" in d:
                    self.optimizer.load_state_dict(d["optimizer_state_dict"])
                if "scheduler_state_dict" in d:
                    self.scheduler.load_state_dict(d["scheduler_state_dict"])
            else:
                # put model in evaluation mode
                self.network.eval()

    def train(
        self,
        train_loader: torch.utils.data.DataLoader,
        test_loader: torch.utils.data.DataLoader,
        train_dir: str,
        runtime_limits: object = None,
        checkpoint_epochs: int = None,
        use_wandb=False,
        test_only=False,
        early_stopping: Optional[EarlyStopping] = None,
    ):
        """

        Parameters
        ----------
        train_loader
        test_loader
        train_dir
        runtime_limits
        checkpoint_epochs
        use_wandb
        test_only: bool = False
            if True, training is skipped
        early_stopping: EarlyStopping
            Optional EarlyStopping instance.

        Returns
        -------

        """

        if test_only:
            test_loss = test_epoch(self, test_loader)
            print(f"test loss: {test_loss:.3f}")

        else:
            while not runtime_limits.limits_exceeded(self.epoch):
                self.epoch += 1

                # Training
                lr = utils.get_lr(self.optimizer)
                with threadpool_limits(limits=1, user_api="blas"):
                    print(f"\nStart training epoch {self.epoch} with lr {lr}")
                    time_start = time.time()
                    train_loss = train_epoch(self, train_loader)
                    train_time = time.time() - time_start

                    print(
                        "Done. This took {:2.0f}:{:2.0f} min.".format(
                            *divmod(train_time, 60)
                        )
                    )

                    # Testing
                    print(f"Start testing epoch {self.epoch}")
                    time_start = time.time()
                    test_loss = test_epoch(self, test_loader)
                    test_time = time.time() - time_start

                    print(
                        "Done. This took {:2.0f}:{:2.0f} min.".format(
                            *divmod(time.time() - time_start, 60)
                        )
                    )

                # scheduler step for learning rate
                utils.perform_scheduler_step(self.scheduler, test_loss)

                # write history and save model
                utils.write_history(train_dir, self.epoch, train_loss, test_loss, lr)
                utils.save_model(self, train_dir, checkpoint_epochs=checkpoint_epochs)
                if use_wandb:
                    try:
                        import wandb

                        wandb.define_metric("epoch")
                        wandb.define_metric("*", step_metric="epoch")
                        wandb.log(
                            {
                                "epoch": self.epoch,
                                "learning_rate": lr[0],
                                "train_loss": train_loss,
                                "test_loss": test_loss,
                                "train_time": train_time,
                                "test_time": test_time,
                            }
                        )
                    except ImportError:
                        print("wandb not installed. Skipping logging to wandb.")

                if early_stopping is not None:
                    # Whether to use train or test loss
                    early_stopping_loss = (
                        test_loss
                        if early_stopping.metric == "validation"
                        else train_loss
                    )
                    is_best_model = early_stopping(early_stopping_loss)
                    if is_best_model:
                        self.save_model(
                            join(train_dir, "best_model.pt"), save_training_info=False
                        )
                    if early_stopping.early_stop:
                        print("Early stopping")
                        break
                print(f"Finished training epoch {self.epoch}.\n")


def train_epoch(pm, dataloader):
    pm.network.train()
    loss_info = dingo.core.utils.trainutils.LossInfo(
        pm.epoch,
        len(dataloader.dataset),
        dataloader.batch_size,
        mode="Train",
        print_freq=1,
    )

    for batch_idx, data in enumerate(dataloader):
        loss_info.update_timer()
        pm.optimizer.zero_grad()
        # data to device
        data = [d.to(pm.device, non_blocking=True) for d in data]
        # compute loss
        loss = pm.loss(data[0], *data[1:])
        # backward pass and optimizer step
        loss.backward()
        pm.optimizer.step()
        # update loss for history and logging
        loss_info.update(loss.detach().item(), len(data[0]))
        loss_info.print_info(batch_idx)

    return loss_info.get_avg()


def test_epoch(pm, dataloader):
    with torch.no_grad():
        pm.network.eval()
        loss_info = dingo.core.utils.trainutils.LossInfo(
            pm.epoch,
            len(dataloader.dataset),
            dataloader.batch_size,
            mode="Test",
            print_freq=1,
        )

        for batch_idx, data in enumerate(dataloader):
            loss_info.update_timer()
            # data to device
            data = [d.to(pm.device, non_blocking=True) for d in data]
            # compute loss
            loss = pm.loss(data[0], *data[1:])
            # update loss for history and logging
            loss_info.update(loss.item(), len(data[0]))
            loss_info.print_info(batch_idx)

        return loss_info.get_avg()



import argparse
import textwrap
from os.path import join

import os
import yaml
import pickle

from dingo.gw.noise.asd_estimation import (
    download_and_estimate_psds,
)
from dingo.gw.noise.asd_dataset import ASDDataset
from dingo.gw.noise.generate_dataset_dag import create_dag
from dingo.gw.noise.utils import merge_datasets, get_time_segments


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """\
        Generate an ASD dataset based on a settings file.
        """
        ),
    )
    parser.add_argument(
        "--data_dir",
        type=str,
        required=True,
        help="Path where the PSD data is to be stored. Must contain a 'settings.yaml' file.",
    )
    parser.add_argument(
        "--settings_file",
        type=str,
        default=None,
        help="Path to a settings file in case two different datasets are generated in the same directory",
    )
    parser.add_argument(
        "--time_segments_file",
        type=str,
        default=None,
        help="Optional file containing a dictionary of a list of time segments that should be used for estimating PSDs."
        "This has to be a pickle file.",
    )
    parser.add_argument(
        "--out_name",
        type=str,
        default=None,
        help="Path to resulting ASD dataset",
    )
    parser.add_argument("--verbose", action="store_true")

    return parser.parse_args()


def generate_dataset():
    """
    Creates and saves an ASD dataset
    """
    args = parse_args()

    # Load settings
    settings_file = (
        args.settings_file
        if args.settings_file is not None
        else join(args.data_dir, "asd_dataset_settings.yaml")
    )
    with open(settings_file, "r") as f:
        settings = yaml.safe_load(f)

    data_dir = args.data_dir

    if args.time_segments_file:
        with open(args.time_segments_file, "rb") as f:
            time_segments = pickle.load(f)
    else:
        time_segments = get_time_segments(settings["dataset_settings"])
        time_segments_path = join(
            data_dir, "tmp", settings["dataset_settings"]["observing_run"]
        )
        os.makedirs(time_segments_path, exist_ok=True)
        with open(join(time_segments_path, "psd_time_segments.pkl"), "wb") as f:
            pickle.dump(time_segments, f)

    if "condor" in settings:

        dagman = create_dag(data_dir, settings_file, time_segments, args.out_name)

        try:
            dagman.visualize(
                join(data_dir, "tmp", "condor", "ASD_dataset_generation_workflow.png")
            )
        except:
            pass

        dagman.build()
        print(f"DAG submission file written.")

    else:

        print("Downloading strain data and estimating PSDs...")
        asd_filename_list = download_and_estimate_psds(
            args.data_dir, settings, time_segments, verbose=args.verbose
        )
        asd_dataset_list = {
            det: [ASDDataset(asd_file) for asd_file in asd_file_list]
            for det, asd_file_list in asd_filename_list.items()
        }
        print("Merging single dataset files into one...")
        dataset = merge_datasets(asd_dataset_list)
        filename = args.out_name
        if filename is None:
            run = settings["dataset_settings"]["observing_run"]
            filename = join(data_dir, f"asds_{run}.hdf5")
        dataset.to_file(filename)
